if( exists("WAME.manual") ) {
	source("estimateSigmaStar.R");
}

WAME.paired<-function(X,design,contrasts,Sigma,alpha,Sigma.star,estimation.method="full nML",cache.dir=getOption("WAME.cache.dir")){
	#TODO: make verifyX
	#TODO: make verifyDesign
	#TODO: split by NA functionality
	#TODO: make verifyTestability

	#If Sigma is given, then it can be used as Sigma.star if we need it when estimating alpha
	if( !missing(Sigma) ) Sigma.star<-Sigma;
	
	#If Sigma.star is not given, then estimate it
	if( missing(Sigma.star) ) {
		Sigma.star<-estimateSigmaStar(X=X,estimation.method=estimation.method,cache.dir=cache.dir);
	}
	
	#If either Sigma or alpha are not given, estimate the scale, lambda, for
	# Sigma.star and alpha and form the estimate of Sigma.
	if( missing(Sigma) || missing(alpha) ) {
		hyper.par<-estimateHyperPar(X,design,Sigma.star);
		alpha<-hyper.par$alpha;
		Sigma<-calcSigma(Sigma.star,hyper.par$lambda);
	}
	
	#Calculate the coefficient-wise weights for the estimation.
	weights=calcWeights(design,contrasts,Sigma);
	
	#Calculate the degrees of freedom like parameters (F1, F2, (t,) n, k, p)
	df<-calcDf(design,contrasts,alpha);
	
	#Calculate the gene-wise estimates of the coefficients (i.e. 'contrasts' times the parameter vector)
	coefficients<-calcCoef(X,weights);
	
	#Calculate the gene-wise variance-related statistic S
	S<-calcS(X,design,Sigma);
	
	#Perform the test, i.e. calculate the gene-wise statistics of interest, as well as the related p-values.
	stats<-calcStats(X,design,contrasts,Sigma,coefficients,S,df);
	
	#Return the results compiled in a list
 	res<-append(stats, list(coefficients=coefficients, design=design, contrasts=contrasts, weights=weights, alpha=alpha, S=S, Sigma=Sigma, df=df));
	class(res)<-"WAME.paired.result";
	return(res);
}

# --- Helper functions --- #

#TODO: make top.table

# estimateSigmaStar is placed in "estimateSigmaStar.R"

estimateHyperPar<-function(X,design,Sigma.star){
	# ---estimateHyperPar---
	#	Maximum likelihood estimates of the hyper-parameters alpha and lambda
	#
	# ---Parameters---
	# X
	#	The data in a gene by repetition matrix.
	# Sigma.star
	#	The scaled covariance matrix
	#
	# ---Result---
	#	The MLEs in a list.
	# ------------

	ngenes<-nrow(X); 
	narrays<-ncol(X); #n
	Drank<-computeRank(design); #p

	# Re-parametrisation functions to be used in the optimisation procedure
	al2r<-function(al){c(al$a,al$l)}; #To represent alpha and lambda as a vector
	r2al<-function(r){list(alpha=r[1],lambda=r[2])};
	
	S<-calcS( X, design, Sigma.star );
	alLL<-function(r){
		a<-r2al(r)$a; l<-r2al(r)$l;
		if( a<=0 || l<=0 ) return( as.double(NA) );
		return( ngenes*((narrays-Drank)/2*log(l)+log(gamma(a+(narrays-Drank)/2))-log(gamma(a)))-(a+(narrays-Drank)/2)*sum(log(S/2*l+1)) );
	}

	start.al=list(a=2,l=1);
	r.tot<-optim(al2r(start.al),alLL,control=list(fnscale=-1,maxit=1000));
	if( r.tot$convergence!=0 ) cat("Potential error: Estimation of alpha and lambda had non-optimal result code (optim): ", r.tot$convergence, "!\n");
	return(r2al(r.tot$par));
}

calcSigma<-function(Sigma.star,lambda){
	# ---calcSigma---
	#	Calculates Sigma from SigmaStar and lambda
	# ------------------
	return( Sigma.star/lambda )
}

#Calculate the coefficient-wise weights for the estimation.
calcWeights<-function(design,contrasts,Sigma){
	iSigma=ginv(Sigma);
	res<-contrasts%*%ginv(t(design)%*%iSigma%*%design)%*%t(design)%*%iSigma;
	colnames(res)<-rownames(design);
	res
}

#Calculate the gene-wise estimates of the coefficients (i.e. 'contrasts' times the parameter vector)
calcCoef<-function(X,weights){
	X%*%t(weights);
}

#Calculate the degrees of freedom like parameters (F1, F2, (t,) n, k, p)
calcDf<-function(design,contrasts,alpha){
	n<-nrow(design); k<-computeRank(contrasts); p<-computeRank(design); F1<-k; F2<-n-p+2*alpha;
	if(F1==1) {
		list( F1=F1,F2=F2,t=F2,n=n,k=k,p=p)
	} else {
		list( F1=F1,F2=F2,n=n,k=k,p=p)
	}
}

#Calculate the gene-wise variance-related statistic S
calcS<-function(X,design,Sigma){
	#Form the projection matrix to the orthogonal complement of V, 
	# i.e. the space spanned by the rows in D
	P.Vc=Pc(D=design,Sigma=Sigma);
	return( norm.sq(X%*%t(P.Vc),Sigma) );
}

#Perform the test, i.e. calculate the gene-wise statistics of interest, as well as the related p-values.
calcStats<-function( 
	X, design, contrasts, Sigma, coefficients,
	S=calcS(X,design,Sigma), df=calcDf(design,contrasts)
){
	#The projection-set contains projections to a variety of spaces, e.g. on V_C. 
	pset<-projection.set(D=design,C=contrasts,Sigma=Sigma);
	F=df$F2/df$F1*norm.sq(X%*%t(pset$P.VC),Sigma)/(S+2);
	p.value=1-pf(q=F, df1=df$F1, df2=df$F2);
	if(df$F1==1) {
		t=sqrt(F)*sign(coefficients);
		return( list(p.value=p.value, F=F, t=t) );
	} else {
		return( list(p.value=p.value, F=F) );
	}
}

# --- Misc functions --- #

# computeRank: compute the (numerical) rank of a matrix
computeRank<-function(M,eps=10^(-5)){
        if( length(M)==0 ) return(0);
        return( sum(abs(svd(M)$d)>eps) )
}

