if( exists("WAME.manual") ) {
	library(MASS);
	library(mgcv);
	source("parametrisations.R");
	source("projectionsAndNorms.R");
}

default.minimizer<-function(f,p){
	nlm(f=f,p=p,iterlim=1000);
}

estimateSigmaStar<-function( 
	X, estimation.method,
	startSigmaStar=Id(ncol(X)),
	minimizer=function(f,p){nlm(f=f,p=p,iterlim=1000)},
	Ps=NULL, cache.dir=getOption("WAME.cache.dir")
){
	# ---estimateSigmaStar---
	#	Estimates the scaled version of Sigma, SigmaStar, given the data.
	#
	# ---Parameters---
	# X
	#	The data in a gene by repetition matrix.
	#TODO: fill in#
	#
	# ---Result---
	# The estimated variance covariance matrix, SigmaStar.
	# ------------
		
	#initialize constants
	ngenes<-nrow(X);
	narrays<-ncol(X);
	
	covPF=make.parametrisation(X=X,estimation.method); #CovarianceParametrisationF
	
	if( is.null(Ps)) {
		return( nMLE.Sigma.Star(covP=covPF(X),X=X,start.cov=startSigmaStar, minimizer=minimizer)$SigmaStar );
	} else {
		space=attributes(covPF)$possible.spaces[1];
		cat("Parametrising the space of ", space, "-values.\n");
		covP=covPF(X,Ps=Ps,space=space); #create the covariance parametrisation (might involve preprocessing estimations, etc)

		n.z=nrow(Ps$x2z);
		if(covP$r.length>(n.z/2*(n.z+1)-1)) cat("FYI: Overparametrisation occurs.", covP$r.length, " parameters for Z of length", n.z, "\n");
		
		#perform the actual numerical ML-estimation in the Z-space
		triple.covP=transformed.covP(covP,Ps,space); #the coupled parametrisations for the three spaces (X,Y,Z)
		start.r=triple.covP$X$cov2r(startSigmaStar); #start with the equivalent of equal var, no corr for X
		r.SigmaStar<-nMLE.Sigma.Star(triple.covP$Z,X=X%*%t(Ps$x2z),start.r=start.r,minimizer=minimizer)$r;
				
		SigmaStars=lapply(triple.covP,
			function(covP){if(!is.null(covP$r2cov)) covP$r2cov(r.SigmaStar) else NULL}
		);
		SigmaStars=SigmaStars[sapply(SigmaStars,function(ss){!is.null(ss)})]; #SigmaStar for X might not be known
		#Note that the SigmaStars have the same scale lambda, and therefore does not necessarily have has 1 in upper left corner
		return(SigmaStars);
	}
}

#nMLE.Sigma.Star: CovarianceParametrisationF -> ScaledCovarianceEstimator (see parametrisations.R)
#numerical ML estimator of SigmaStar, given the parametrisation.
nMLE.Sigma.Star<-function(covP,X,start.cov=Id(ncol(X)),start.r,minimizer=function(f,p){nlm(f=f,p=p,iterlim=1000)}){
	narrays=ncol(X);
	ngenes=nrow(X);
	
	#get the start-value for the optimisation
	if(missing(start.r)) start.r<-covP$cov2r(start.cov);
	
	if( length(start.r)==0 ) {
		#sometimes nothing needs to be estimated, e.g. when no correlations and equal variances are specified
		SigmaStar<-covP$r2cov(start.r); r.SigmaStar=start.r;
	} else {
		#create the negative log-likelihood function	
		V=computeV(X); #compute V once, as needed by the likelihood function in each iteration
		neval<-0;
		negLL<-function(r) {
			# ---negLL---
			#	-1 * the log-likelihood function for SigmaStar given the data.
			#	*-1 is to make it a minimization problem instead of a maximization problem, which is what nlm solves.
			# -----------
					
			SigmaS<-covP$r2cov(r);
			if( any(is.nan(SigmaS)) || max(abs(SigmaS))==Inf || min(eigen(SigmaS,only.values=T)$values)<0 || min(diag(SigmaS))<0 ) {
				return(as.double(NA));
			}
			
			#increase the evaluation counter
			neval<<-neval+1;
			
			res<-ngenes*log(det(SigmaS))+narrays*sum(log(norm.sq(V,Sigma=SigmaS)));
			return( res );
		}

		#perform the actual numerical maximisation of the likelihood
		r.tot<-minimizer(f=negLL,p=start.r);
		if( r.tot$code>2 ) cat("Potential error: Estimation of SigmaStar had non-optimal result code (nlm): ", r.tot$code, "!\n");
		SigmaStar<-covP$r2cov(r.tot$estimate);
		r.SigmaStar=r.tot$estimate;
		cat("Iterations:", r.tot$iter, "Evaluations:", neval, "Maximal LL:",-r.tot$minimum,"\n");
	}
	dimnames(SigmaStar)<-list(colnames(X),colnames(X)); #keep the names of the repetitions (arrays)	
	return( list(SigmaStar=SigmaStar,r=r.SigmaStar) );
}

computeV <- function(X) {
	# ---computeV---
	#	Performs a transformation of the data, to cancel out c_g.
	# --------------
	
	X[X[,1]==0,1]<-0.00001; #to remove problem with division by zero
	return(X/X[,1]);
}
