##AS060316: below needs update
## The parametrisation is based on the following types of functions
# CorrelationParametrisation: list(cor2r,r2cor,r.length)
# CorrelationParametrisationF: Data -> CorrelationParametrisation
# preCorrelationParametrisationF: (...,space) -> CorrelationParametrisationF
# CovarianceParametrisation: list(cov2r,r2cov,r.length)
# CovarianceParametrisationF: Data -> CovarianceParametrisation
# ScaledCovarianceEstimator: Data -> CovarianceMatrix
# Data : (X[,Ps,spaces])
#
# P is a set of projections, which are included in an experimental generalisation of paired WAME.
# spaces is the set of spaces where the estimation is allowed takes place (X, Y, Z)
# P and spaces will be left out for the currently documented case
#
#
# Here, cov/r2r and r2cov/r are the functions that convert back and forth 
# between the parametrisation that the optimiser uses and the covariance / correlation matrix.
# X is the data that the parametrisation is to be done for. Typically a full parametrisation
# only uses ncol(X).

#----------------------------------#
#---Correlation Parametrisations---#
#----------------------------------#

#freeCor: CorrelationParametrisationF
rawFreeCorPF<-function(X){
	narrays=ncol(X); rm("X");
	r.length=narrays*(narrays-1)/2;
	list( 
		cor2r=function(M){M[lower.tri(M)]}, 
		r2cor=function(r){
			M<-matrix(0,ncol=narrays,nrow=narrays);
			M[lower.tri(M)]<-r; #take care of the lower triangle
			M<-M+t(M)+Id(narrays); #transpose to the upper triangle and fill the diagonal with ones
			M
		},
		r.length=r.length
	);
}
freeCor<-function(space=c("Z","Y","X")){as.corPF(rawFreeCorPF,possible.spaces=space)};
#freeCor may parametrise any of the spaces

#medianEstimatedCor: CorrelationParametrisationF
rawMedianEstimatedCorPF<-function(X){
	estCor=estimateCorM.median(X); estCor; rm("X");
	list( cor2r=function(M){c()}, r2cor=function(r){estCor}, r.length=0);
}
medianEstimatedCor<-function(space=c("Y","Z")){as.corPF(rawMedianEstimatedCorPF,possible.spaces=space)};
#medianEstimatedCor can only estimate correlations in "Z" and "Y" spaces

#fixedCor: (CorrelationMatrix,Space)->CorrelationParametrisationF
rawFixedCor<-function(corM){ function(X){ #CorrelationParametrisationF
	rm("X");
	list( cor2r=function(M){c()}, r2cor=function(r){corM}, r.length=0 );
}}
fixedCor<-function(corM,space="X"){as.corPF(rawFixedCor(corM),possible.spaces=space)}

#noCor: Space->CorrelationParametrisationF
rawNoCorPF<-function(X){
	narrays=ncol(X); rm("X");
	corM=Id(narrays);
	list( cor2r=function(M){c()}, r2cor=function(r){corM}, r.length=0 );
}
noCor<-function(space="X"){as.corPF(rawNoCorPF,possible.spaces=space)}

#specCor: (CorrelationStructure,Space)->CorrelationParametrisationF
rawSpecCorPF<-function(corStruct){
	corStruct<-as.corStructCompS(corStruct);
	if(length(corStruct)==0) return( rawNoCorPF );

	function(X){ #the CorrelationParametrisationF
		rm("X");
		list(
			cor2r=function(M){
				#find one element in M for each list element, set the respective r to that
				sapply(corStruct,function(B){M[B][1]});
			},
			r2cor=function(r){
				#the weights for the components are given in r
				#perform a weighted sum of the list elements
				M<-corStruct[[1]]*0; #gets the right size and dim.names
				for( i in 1:length(corStruct) ) {
					M<-M+corStruct[[i]]*r[i];
				}
				diag(M)=1;
				M
			},
			r.length=length(corStruct)
		)
	}
}
specCor<-function(corStruct,space="X"){as.corPF(rawSpecCorPF(corStruct),possible.spaces=space)}

#-------------------------------#
#---Variance Parametrisations---#
#-------------------------------#

#rawFreeVar : rawVarianceParametrisationF
rawFreeVarPF<-function(X){
	narrays=ncol(X); rm("X");
	list(
		var2r=function(V){V[-1]}, 
		r2var=function(r){c(1,r)},
		r.length=narrays-1
	);	
}
freeVar<-function(space=c("Z","Y","X")){as.varPF(rawFreeVarPF,possible.spaces=space)};

#specVar:(VarianceStructure,Space)->CovarianceParametrisationF
rawSpecVar<-function(varStruct){ 
	varStruct<-as.varStructCompS(varStruct);

	#find the component parametrising the upper left variance and put it first in the list
	ul<-which(sapply(varStruct,function(B){B[1]})); 
	varStruct=append(varStruct[ul],varStruct[-ul]);
	n.comp=length(varStruct); #the number of components (including the first)

	function(X){
		n=ncol(X);
		list( #uses n-1 parameters, since first variance is one
			var2r=function(V) {
				#find one element in V for each list element, set the respective r to that
				#the upper left elements is always one, so disregard it
				sapply(varStruct,function(B){V[B][1]})[-1];
			},
			r2var=function(r){
				r<-c(1,r); #the first element has weight 1

				#perform a weighted sum of the list elements
				V<-double(n);
				for( i in 1:n.comp ) {
					V<-V+varStruct[[i]]*r[i];
				}
				V
			},
			r.length=n.comp-1
		);
	}
}
specVar<-function(varStruct,space="X"){as.varPF(rawSpecVar(varStruct),possible.spaces=space)}

#---------------------------------#
#---Covariance Parametrisations---#
#---------------------------------#

#freeCov: CovarianceParametrisationF
freeCov<-function(space=c("Z","Y","X")){as.covPF(var=freeVar(space),cor=freeCor(space))};

#specCov: (CovarianceStructureMatrix,Space)->CovarianceParametrisationF
specCov<-function(M,space="X"){as.covPF(var=specVar(M,space),cor=specCor(M,space))};

#---------------------#
#---Other functions---#
#---------------------#

#make.parametrisation - produces covPF from a number of inputs
make.parametrisation<-function(X, estimation.method){
	#Determine what method is wanted
	if( class(estimation.method)=="CovarianceParametrisationF" ) {
		return(estimation.method);
	}

	if( is.matrix(estimation.method) ){
		return(specCov(estimation.method));
	}
	
	if( is.character(estimation.method) ) {
		if( any(grep(estimation.method,"median corr + nML",extended=F)) ) {
			estimation.method=list(freeVar(),medianEstimatedCor());
		} else {
			if( !any(grep(estimation.method,"full nML",extended=F)) ) {
				warning("invalid estimation.method:", estimation.method, ". Using \"full nML\".");
			}
			return(freeCov()); #"full nML";
		}
	}

	if( is.list(estimation.method) ) {
		corPF.I<-which(sapply(estimation.method, function(x){class(x)=="CorrelationParametrisationF"}))
		varPF.I<-which(sapply(estimation.method, function(x){class(x)=="VarianceParametrisationF"}))
		if( length(corPF.I)>0 && length(varPF.I>0) ) {
			return(as.covPF(cor=estimation.method[[corPF.I[1]]],var=estimation.method[[varPF.I[1]]]))
		}
	}
	
	stop( "invalid estimation.method" );
}

#-----------------------#
#---Wrapper functions---#
#-----------------------#

#the wrappers take care of common tasks, such as setting the right class of the result
#[experimental: as well as communication between the covPF and corPF when deciding which
# space the estiamtion will be performed in]

as.covPF<-function(corPF,varPF){
	possible.spaces=attributes(corPF)$possible.spaces[attributes(corPF)$possible.spaces%in%attributes(varPF)$possible.spaces];

	covPF<-function(X,Ps,space=possible.spaces[1]){ 		
		varP<-varPF(X,Ps,space);
		corP<-corPF(X,Ps,space);
		r.length.var=varP$r.length;
		r.length.cor=corP$r.length;
		if(missing(Ps)) Ps<-NULL;
		
		covP<-list(
			cov2r=function(M){
				r.var=varP$var2r(diag(M));
				r.cor=corP$cor2r(cov2Cor(M));
				c(r.var,r.cor);
			},
			r2cov=function(r){
				r.var<-r[(1:(r.length.var+1))[-(r.length.var+1)]]; #the last part to handle length 0; 1:0 != c()
				r.cor<-r[(r.length.var:length(r))[-1]]; #the last part to handle length 0
				v<-varP$r2var(r.var);
				cM<-corP$r2cor(r.cor);
				res<-diag(sqrt(v))%*%cM%*%diag(sqrt(v));
				if(!is.null(Ps) &&  space=="Y") res<-Ps$x2y%*%res%*%t(Ps$x2y); #just to make sure the cov matrix is in the right space
				res;
			},
			r.length=r.length.var+r.length.cor
		);
		class(covP)<-"CovarianceParametrisation"
		covP;
	}

	attributes(covPF)<-append(attributes(covPF),list(possible.spaces=possible.spaces));
	class(covPF)<-"CovarianceParametrisationF";
	covPF
}

as.corPF<-function(rawCorPF,possible.spaces=c("Z","Y","X")){
	corPF<-function(X,Ps,space){
		if( missing(Ps) || is.null(Ps) ) {
			corP<-rawCorPF(X);
		} else {
			if( !space%in%possible.spaces ) stop("Selected space (",space,") not among possible ones (", paste(possible.spaces,collapse=","),")");
			Ps2<-list(X=Id(ncol(X)),Y=Ps$x2y,Z=Ps$x2z); #just renaming to suite the apply below
			corP<-rawCorPF(X%*%t(Ps2[[space]]));
			attributes(corP)<-append(attributes(corP),list(space=space));
		}
		class(corP)<-"CorrelationParametrisation";
		return(corP);	
	}
	class(corPF)<-"CorrelationParametrisationF";
	attributes(corPF)<-append(attributes(corPF),list(possible.spaces=possible.spaces));
	return(corPF);
}

as.varPF<-function(rawVarPF,possible.spaces=c("Z","Y","X")){
	varPF<-function(X,Ps,space){
		if( missing(Ps) || is.null(Ps) ) {
			varP<-rawVarPF(X);
		} else {
			if( !space%in%possible.spaces ) stop("Selected space (",space,") not among possible ones (", paste(possible.spaces,collapse=","),")");
			Ps2<-list(X=Id(ncol(X)),Y=Ps$x2y,Z=Ps$x2z); #just renaming to suite the apply below
			varP<-rawVarPF(X%*%t(Ps2[[space]]));
			attributes(varP)<-append(attributes(varP),list(space=space));
		}
		class(varP)<-"VarianceParametrisation";
		return(varP);	
	}
	class(varPF)<-"VarianceParametrisationF";
	attributes(varPF)<-append(attributes(varPF),list(possible.spaces=possible.spaces));
	return(varPF);
}

# transformed.covP: given a parametrisation for the covariance matrix in one space, 
# give the coupled parametrisations of the other spaces (when feasible).
transformed.covP<-function(covP,Ps,space){
	if( space=="X" ) {
		covP.X=covP;
		covP.Y=list(r2cov=function(r){Ps$x2y%*%covP$r2cov(r)%*%t(Ps$x2y)},r.length=covP$r.length);
		covP.Z=list(r2cov=function(r){Ps$x2z%*%covP$r2cov(r)%*%t(Ps$x2z)},r.length=covP$r.length);
	}
	if( space=="Y" ) {
		covP.Y=covP;
		covP.X=list(cov2r=function(M){covP$cov2r(Ps$x2y%*%M%*%t(Ps$x2y))});
		covP.Z=list(cov2r=function(M){covP$cov2r(Ps$z2y%*%M%*%t(Ps$z2y))},
		            r2cov=function(r){Ps$y2z%*%covP$r2cov(r)%*%t(Ps$y2z)},
					r.length=covP$r.length);
	}
	if( space=="Z" ) {
		covP.Z=covP;
		covP.X=list(cov2r=function(M){covP$cov2r(Ps$x2z%*%M%*%t(Ps$x2z))});
		covP.Y=list(cov2r=function(M){covP$cov2r(Ps$y2z%*%M%*%t(Ps$y2z))},
		            r2cov=function(r){Ps$z2y%*%covP$r2cov(r)%*%t(Ps$z2y)},
					r.length=covP$r.length);
	}
	#class(covP.X)=class(covP.Y)=class(covP.Z)="CovarianceParametrisation";
	covPl<-list(X=covP.X,Y=covP.Y,Z=covP.Z);
	covPl
}

#----------------------#
#---Helper functions---#
#----------------------#

estimateCorM.median <- function( X ) {
#helper function
#based on that median(Xi/Xj)*median(Xj/Xi)=rho_ij^2
	n<-ncol(X);
	r<-matrix(1,ncol=n,nrow=n,dimnames=list(colnames(X),colnames(X)));
	for( i in 1:(n-1) ) for( j in (i+1):n ) {
		r1<-median(X[,i]/X[,j]);
		r2<-median(X[,j]/X[,i]);
		r[j,i]<-r[i,j]<-sqrt(abs(r1*r2))*sign(r1);
	}
	return(r)
}

# as.corStructCompS : makes correlation structure components, 
# i.e. a list of boolean matrices, indicating which elements in the correlation matrix are equal
as.corStructCompS<-function(M){
	if(class(M)=="CorrelationStructureComponents") return(M);

	diag(M)=0; #the diagonal is non-coding

	#make sure the matrix is symmetric
	M[upper.tri(M)]<-0;
	M<-M+t(M);
	
	valS<-setdiff(unique(as.vector(M)),0);
	Bs<-lapply(valS,function(val){M==val});
	class(Bs)<-"CorrelationStructureComponents";
	Bs
}

as.varStructCompS<-function(V){
	if(class(V)=="CovarianceStructureComponents") return(V);

	if(is.matrix(V)) V<-diag(V); #only the diagonal is coding
	valS<-unique(V);
	Bs<-lapply(valS,function(val){V==val});
	class(Bs)<-"CovarianceStructureComponents";
	Bs
}

#cov2Cor : covariance matrix to correlation matrix
cov2Cor<-function(Sigma) {
	res=diag(1/sqrt(diag(Sigma)))%*%Sigma%*%diag(1/sqrt(diag(Sigma)));
	dimnames(res)=dimnames(Sigma);
	diag(res)=1; #make it exact 1
	res
}
